//======================================================================
// Driver for the Epson RTC module RX-4803 SA
// 
// Copyright(C) SEIKO EPSON CORPORATION 2014. All rights reserved.
// 
// Derived from RX-8025 driver: 
// Copyright (C) 2009 Wolfgang Grandegger <wg@grandegger.com>
// 
// Copyright (C) 2005 by Digi International Inc.
// All rights reserved.
// 
// Modified by fengjh at rising.com.cn
// <http://lists.lm-sensors.org/mailman/listinfo/lm-sensors>
// 2006.11
// 
// Code cleanup by Sergei Poselenov, <sposelenov@emcraft.com>
// Converted to new style by Wolfgang Grandegger <wg@grandegger.com>
// Alarm and periodic interrupt added by Dmitry Rakhchev <rda@emcraft.com>
//
// 
// This driver software is distributed as is, without any warranty of any kind,
// either express or implied as further specified in the GNU Public License. This
// software may be used and distributed according to the terms of the GNU Public
// License, version 2 as published by the Free Software Foundation.
// See the file COPYING in the main directory of this archive for more details.
//
// You should have received a copy of the GNU General Public License along with
// this program. If not, see <http://www.gnu.org/licenses/>.
//======================================================================

#if 0
#define DEBUG
#include <linux/device.h>
#undef DEBUG
#endif 

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/bcd.h>
#include <linux/spi/spi.h>
#include <linux/list.h>
#include <linux/rtc.h>
#include <linux/of_gpio.h>

#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/of_irq.h>
#include <linux/interrupt.h>
#include <linux/input.h>
	

// Bank 1, RX-4803 Basic Time and Calendar Register definitions
#define RX4803_BTC_SEC					0x10
#define RX4803_BTC_MIN					0x11
#define RX4803_BTC_HOUR					0x12
#define RX4803_BTC_WEEK					0x13
#define RX4803_BTC_DAY					0x14
#define RX4803_BTC_MONTH				0x15
#define RX4803_BTC_YEAR					0x16
#define RX4803_BTC_RAM					0x17
#define RX4803_BTC_ALARM_MIN			0x18
#define RX4803_BTC_ALARM_HOUR			0x19
#define RX4803_BTC_ALARM_WEEK_OR_DAY	0x1A
#define RX4803_BTC_TIMER_CNT_0			0x1B
#define RX4803_BTC_TIMER_CNT_1			0x1C
#define RX4803_BTC_EXT					0x1D
#define RX4803_BTC_FLAG					0x1E
#define RX4803_BTC_CTRL					0x1F

// Bank 2, RX-4803 Extension Register 1 definitions
#define RX4803_EX1_100S					0x20
#define RX4803_EX1_SEC					0x21
#define RX4803_EX1_MIN					0x22
#define RX4803_EX1_HOUR					0x23
#define RX4803_EX1_WEEK					0x24
#define RX4803_EX1_DAY					0x25
#define RX4803_EX1_MONTH				0x26
#define RX4803_EX1_YEAR					0x27
#define RX4803_EX1_ALARM_MIN			0x28
#define RX4803_EX1_ALARM_HOUR			0x29
#define RX4803_EX1_ALARM_WEEK_OR_DAY	0x2A
#define RX4803_EX1_TIMER_CNT_0			0x2B
#define RX4803_EX1_TIMER_CNT_1			0x2C
#define RX4803_EX1_EXT					0x2D
#define RX4803_EX1_FLAG					0x2E
#define RX4803_EX1_CTRL					0x2F

// Bank 3, RX-4803 Extension Register 2 definitions
#define RX4803_EX2_100S_CP				0x30
#define RX4803_EX2_SEC_CP				0x31

#define RX4803_EX2_OSC_OFFSET			0x3C

#define RX4803_EX2_EVENT_CTRL			0x3F

// Flag RX4803_BTC_EXT Register bit positions
#define RX4803_BTC_EXT_TSEL0		(1 << 0)
#define RX4803_BTC_EXT_TSEL1		(1 << 1)
#define RX4803_BTC_EXT_FSEL0		(1 << 2)
#define RX4803_BTC_EXT_FSEL1		(1 << 3) 
#define RX4803_BTC_EXT_TE 			(1 << 4)
#define RX4803_BTC_EXT_USEL			(1 << 5) 
#define RX4803_BTC_EXT_WADA			(1 << 6)
#define RX4803_BTC_EXT_TEST			(1 << 7)

// Flag RX4803_BTC_FLAG Register bit positions
#define RX4803_BTC_FLAG_VDET 		(1 << 0)
#define RX4803_BTC_FLAG_VLF 		(1 << 1)
#define RX4803_BTC_FLAG_EVF			(1 << 2)
#define RX4803_BTC_FLAG_AF 			(1 << 3)
#define RX4803_BTC_FLAG_TF 			(1 << 4)
#define RX4803_BTC_FLAG_UF 			(1 << 5)

// Flag RX4803_BTC_FLAG Register bit positions
#define RX4803_BTC_CTRL_RESET 		(1 << 0)

#define RX4803_BTC_CTRL_EIE			(1 << 2)
#define RX4803_BTC_CTRL_AIE 		(1 << 3)
#define RX4803_BTC_CTRL_TIE 		(1 << 4)
#define RX4803_BTC_CTRL_UIE 		(1 << 5)
#define RX4803_BTC_CTRL_CSEL0 		(1 << 6)
#define RX4803_BTC_CTRL_CSEL1		(1 << 7)


static struct spi_driver rx4803_driver;

static const struct spi_device_id rx4803_id[] = {
	{ "rx4803", 0 },
	{ }
};
MODULE_DEVICE_TABLE(spi, rx4803_id);

struct rx4803_data {
	struct spi_device *spi;
	struct rtc_device *rtc;
	struct work_struct work;
	u8 ctrlreg;
	unsigned exiting:1;
};

typedef struct {
	u8 number;
	u8 value;
}reg_data;

#define SE_RTC_REG_READ		_IOWR('p', 0x20, reg_data)		
#define SE_RTC_REG_WRITE	_IOW('p',  0x21, reg_data)	

//----------------------------------------------------------------------
// rx4803_read_reg()
// reads a rx4803 register (see Register defines)
//
//----------------------------------------------------------------------
static int rx4803_read_reg(struct spi_device *spi, u8 address, u8 *data)
{
	int ret;
	
	*data = address | 0x80;
	ret = spi_write_then_read(spi, data, 1, data, 1);	
	if (ret < 0) {
		dev_err(&spi->dev, "Unable to read register #%d\n", address);
		return ret;
	}
	dev_dbg(&spi->dev, "%s: REG[0x%02x] => 0x%02x\n", __func__, address, data[0]);

	return 0;
}

//----------------------------------------------------------------------
// rx4803_write_reg()
// writes a rx4803 register (see Register defines)
//
//----------------------------------------------------------------------
static int rx4803_write_reg(struct spi_device *spi, u8 address, u8 data)
{
	int ret;
	u8 buf[2];

	buf[0] = address & 0x7f;
	buf[1] = data;

	ret = spi_write(spi, buf, 2);	
	dev_dbg(&spi->dev, "%s: REG[0x%02x] <= 0x%02x\n", __func__, buf[0], buf[1]);
	
	if (ret)
		dev_err(&spi->dev, "Unable to write register #%d\n", address);

	return ret;	
}

//----------------------------------------------------------------------
// rx4803_irq()
// irq handler
//
//----------------------------------------------------------------------
static irqreturn_t rx4803_irq(int irq, void *p)
{
	struct rx4803_data *rx4803 = dev_get_drvdata(p);

	disable_irq_nosync(irq);
	schedule_work(&rx4803->work);

	return IRQ_HANDLED;
}

//----------------------------------------------------------------------
// rx4803_work()
//
//----------------------------------------------------------------------
static void rx4803_work(struct work_struct *work)
{
	struct rx4803_data *rx4803 = container_of(work, struct rx4803_data, work);
	struct mutex *lock = &rx4803->rtc->ops_lock;
	struct spi_device *spi = rx4803->spi;
	u8 flags;

	mutex_lock(lock);
	
	if (rx4803_read_reg(spi, RX4803_BTC_FLAG, &flags))
		goto out;
		
	dev_dbg(&spi->dev, "%s REG[%02xh]=>%02xh\n", __func__, RX4803_BTC_FLAG, flags);

	if (flags & RX4803_BTC_FLAG_VLF)
		dev_warn(&spi->dev, "Data loss is detected. All registers must be initialized.\n");
		
	if (flags & RX4803_BTC_FLAG_VDET)
		dev_warn(&spi->dev, "Temperature compensation stop detected.\n");		

	// fixed-cycle timer
	if (flags & RX4803_BTC_FLAG_TF) { 
		flags &= ~RX4803_BTC_FLAG_TF; 
		local_irq_disable();
		rtc_update_irq(rx4803->rtc, 1, RTC_PF);// | RTC_IRQF);
		local_irq_enable();
		dev_dbg(&spi->dev, "%s: fixed-cycle timer function status: %xh\n", __func__, flags);
	}
	
	// alarm function
	if (flags & RX4803_BTC_FLAG_AF) { 
		flags &= ~RX4803_BTC_FLAG_AF; 
		local_irq_disable();
		rtc_update_irq(rx4803->rtc, 1, RTC_AF);// | RTC_IRQF);
		local_irq_enable();
		dev_dbg(&spi->dev, "%s: alarm function status: %xh\n", __func__, flags);
	}
	
	// time update function
	if (flags & RX4803_BTC_FLAG_UF) { 
		flags &= ~RX4803_BTC_FLAG_UF; 
		local_irq_disable();
		rtc_update_irq(rx4803->rtc, 1, RTC_UF);// | RTC_IRQF);
		local_irq_enable();
		dev_dbg(&spi->dev, "%s: time update function status: %xh\n", __func__, flags);
	}

	// acknowledge IRQ
	rx4803_write_reg(spi, RX4803_BTC_FLAG, 0x0f & flags);	

out:
	if (!rx4803->exiting)
		enable_irq(spi->irq);

	mutex_unlock(lock);
}

static int rx4803_get_week_day( u8 reg_week_day )
{
	int i, tm_wday = -1;
	
	for ( i=0; i < 7; i++ )
	{
		if ( reg_week_day & 1 )
		{
			tm_wday = i;
			break;
		}
		reg_week_day >>= 1;
	}
	
	return 	tm_wday;
}

//----------------------------------------------------------------------
// rx4803_get_time()
// gets the current time from the rx4803 registers
//
//----------------------------------------------------------------------
static int rx4803_get_time(struct device *dev, struct rtc_time *dt)
{
	struct spi_device *spi = to_spi_device(dev); 
	u8 date[7];
	int err;
	
	date[0] = RX4803_BTC_SEC | 0x80;
	err = spi_write_then_read(spi, date, 1, date, 7);	
	if (err < 0) {
		dev_err(&spi->dev, "Unable to read registers from #%d\n", RX4803_BTC_SEC);
		return err;
	}

	dev_dbg(dev, "%s: read 0x%02x 0x%02x "
		"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n", __func__,
		date[0], date[1], date[2], date[3], date[4], date[5], date[6]);

	dt->tm_sec  = bcd2bin(date[RX4803_BTC_SEC   - 0x10] & 0x7f);
	dt->tm_min  = bcd2bin(date[RX4803_BTC_MIN   - 0x10] & 0x7f);
	dt->tm_hour = bcd2bin(date[RX4803_BTC_HOUR  - 0x10] & 0x3f);
	dt->tm_wday = rx4803_get_week_day( date[RX4803_BTC_WEEK - 0x10] & 0x7f );
	dt->tm_mday = bcd2bin(date[RX4803_BTC_DAY   - 0x10] & 0x3f);
	dt->tm_mon  = bcd2bin(date[RX4803_BTC_MONTH - 0x10] & 0x1f) - 1;
	dt->tm_year = bcd2bin(date[RX4803_BTC_YEAR  - 0x10]);
	
	if (dt->tm_year < 70)
		dt->tm_year += 100;

	dev_dbg(dev, "%s: date %ds %dm %dh %dwd %dmd %dm %dy\n", __func__,
		dt->tm_sec, dt->tm_min, dt->tm_hour, dt->tm_wday,
		dt->tm_mday, dt->tm_mon, dt->tm_year);

	return rtc_valid_tm(dt);
}

//----------------------------------------------------------------------
// rx4803_set_time()
// Sets the current time in the rx4803 registers
//
//----------------------------------------------------------------------
static int rx4803_set_time(struct device *dev, struct rtc_time *dt)
{
	struct rx4803_data *rx4803 = dev_get_drvdata(dev);
	struct spi_device *spi = rx4803->spi; 
	u8 date[8] = { RX4803_BTC_SEC, };
	int ret = 0;
	
	date[RX4803_BTC_SEC   - 0x10 + 1] = bin2bcd(dt->tm_sec);
	date[RX4803_BTC_MIN   - 0x10 + 1] = bin2bcd(dt->tm_min);
	date[RX4803_BTC_HOUR  - 0x10 + 1] = bin2bcd(dt->tm_hour);		
	date[RX4803_BTC_WEEK  - 0x10 + 1] = 1 << (dt->tm_wday);
	date[RX4803_BTC_DAY   - 0x10 + 1] = bin2bcd(dt->tm_mday);
	date[RX4803_BTC_MONTH - 0x10 + 1] = bin2bcd(dt->tm_mon + 1);
	date[RX4803_BTC_YEAR  - 0x10 + 1] = bin2bcd(dt->tm_year % 100);

	dev_dbg(dev, "%s: write 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
		__func__, date[0+1], date[1+1], date[2+1], date[3+1], date[4+1], date[5+1], date[6+1]);

	ret =  spi_write(spi, date, 8);	
	if (ret)
		dev_err(&spi->dev, "Unable to write register #%d\n", date[0]);

	return ret;
}

//----------------------------------------------------------------------
// rx4803_init()
// initializes the rx4803
//
//----------------------------------------------------------------------
static int rx4803_init(struct spi_device *spi, int *need_reset)
{
	u8 flags;
	int need_clear = 0;
	int err = 0;

	err = rx4803_read_reg(spi, RX4803_BTC_FLAG, &flags);
	if (err)
		goto out;
		
	if ( flags & RX4803_BTC_FLAG_VDET )
	{
		dev_warn(&spi->dev, "Temperature compensation is stop detected.\n");
		need_clear = 1;		
	}
	
	if ( flags & RX4803_BTC_FLAG_VLF )
	{
		dev_warn(&spi->dev, "Data loss is detected. All registers must be initialized.\n");
		*need_reset = 1;	
		need_clear = 1;	
	}
		
	
	if ( flags & RX4803_BTC_FLAG_EVF ){
		dev_warn(&spi->dev, "Event was detected\n");
		need_clear = 1;
	}		
	
	if ( flags & RX4803_BTC_FLAG_AF ){
		dev_warn(&spi->dev, "Alarm was detected\n");
		need_clear = 1;
	}	
	
	if ( flags & RX4803_BTC_FLAG_TF ){
		dev_warn(&spi->dev, "Timer was detected\n");
		need_clear = 1;
	}	
	
	if ( flags & RX4803_BTC_FLAG_UF ){
		dev_warn(&spi->dev, "Update was detected\n");
		need_clear = 1;
	}	

	
	if (*need_reset ) {
		//clear ctrl register
		err = rx4803_write_reg(spi, RX4803_BTC_CTRL, RX4803_BTC_CTRL_CSEL0);
		if (err)
			goto out;	
		
		//set second update
		err = rx4803_write_reg(spi, RX4803_BTC_EXT, 0x00);
		if (err)
			goto out;
			
	}
	
	if (need_clear) {
		//clear flag register
		err = rx4803_write_reg(spi, RX4803_BTC_FLAG, 0x00);
		if (err)
			goto out;
		
	}	

out:
	return err;

}

//----------------------------------------------------------------------
// rx4803_get_alarm()
// reads current Alarm 
//
//----------------------------------------------------------------------
static int rx4803_get_alarm(struct device *dev, struct rtc_wkalrm *t)
{
	struct rx4803_data *rx4803 = dev_get_drvdata(dev);
	struct spi_device *spi = rx4803->spi;
	u8 alarmvals[3];		//minute, hour, week/day values
	u8 ctrl[2];				//extension, flag
	int err;

	if (spi->irq <= 0)
		return -EINVAL;

	//get current minute, hour, week/day alarm values
	alarmvals[0] = RX4803_BTC_ALARM_MIN | 0x80;
	err = spi_write_then_read(spi, alarmvals, 1, alarmvals, 3);	
	if (err < 0) {
		dev_err(&spi->dev, "Unable to read registers from #%d\n", RX4803_BTC_ALARM_MIN);
		return err;
	}		
		
	//get current extension, flag, control values
	ctrl[0] = RX4803_BTC_EXT | 0x80;
	err = spi_write_then_read(spi, ctrl, 1, ctrl, 3);	
	if (err < 0) {
		dev_err(&spi->dev, "Unable to read registers from #%d\n", RX4803_BTC_EXT);
		return err;
	}				
	
	dev_dbg(dev, "%s: minutes:0x%02x hours:0x%02x week/day:0x%02x\n",
		__func__, alarmvals[0], alarmvals[1], alarmvals[2]);

	// Hardware alarm precision is 1 minute
	t->time.tm_sec  = 0;
	t->time.tm_min  = bcd2bin(alarmvals[0] & 0x7f);		
	t->time.tm_hour = bcd2bin(alarmvals[1] & 0x3f);		
		
	if ( ctrl[0] & RX4803_BTC_EXT_WADA )
	{   // Day Alarm
		t->time.tm_wday = -1;
		t->time.tm_mday = bcd2bin(alarmvals[2] & 0x3f);	
	}
	else
	{   // Week Day Alarm
		t->time.tm_wday = rx4803_get_week_day( alarmvals[2] & 0x7f );
		t->time.tm_mday = -1;
	}
	
	t->time.tm_mon  = -1;
	t->time.tm_year = -1;

	dev_dbg(dev, "%s: date: %ds %dm %dh %dwd %dmd %dm %dy\n",
		__func__,
		t->time.tm_sec, t->time.tm_min, t->time.tm_hour, t->time.tm_wday,
		t->time.tm_mday, t->time.tm_mon, t->time.tm_year);
	
	//check if INTR is enabled
	t->enabled = !!(rx4803->ctrlreg & RX4803_BTC_CTRL_AIE);	
	//check if flag is triggered 
	t->pending = (ctrl[1] & RX4803_BTC_FLAG_AF) && t->enabled; 
	
	dev_dbg(dev, "%s: t->enabled: %d t->pending: %d\n",
		__func__,
		t->enabled, t->pending);

	return err;
}

//----------------------------------------------------------------------
// rx4803_set_alarm()
// sets Alarm 
//
//----------------------------------------------------------------------
static int rx4803_set_alarm(struct device *dev, struct rtc_wkalrm *t)
{
	struct rx4803_data *rx4803 = dev_get_drvdata(dev);
	struct spi_device *spi = rx4803->spi;
	u8 alarmvals[4];		//addr, minute, hour, day
	u8 ctrl[2];				//ext, flag registers
	int err;
 
	if (spi->irq <= 0)
		return -EINVAL;
			
	dev_dbg(dev, "%s: date: %ds %dm %dh %dwd %dmd %dm %dy\n",
		__func__,
		t->time.tm_sec, t->time.tm_min, t->time.tm_hour, t->time.tm_wday,
		t->time.tm_mday, t->time.tm_mon, t->time.tm_year);
			
	//get current flag register
	ctrl[0] = RX4803_BTC_FLAG | 0x80;
	err = spi_write_then_read(spi, ctrl, 1, ctrl, 2);	
	if (err < 0) {
		dev_err(&spi->dev, "Unable to read registers from #%d\n", RX4803_BTC_FLAG);
		return err;
	}	

	// Hardware alarm precision is 1 minute
	alarmvals[0+1] = bin2bcd(t->time.tm_min);
	alarmvals[1+1] = bin2bcd(t->time.tm_hour);
	
	if ( ctrl[0] & RX4803_BTC_EXT_WADA )
	{   // Day Alarm
		alarmvals[2+1] = bin2bcd(t->time.tm_mday);
		dev_dbg(dev, "%s: write min:0x%02x hour:0x%02x mday:0x%02x\n", 
		__func__, alarmvals[0+1], alarmvals[1+1], alarmvals[2+1]);
	}
	else
	{   // Week Day Alarm
		alarmvals[2+1] = 1 << (t->time.tm_wday);
		dev_dbg(dev, "%s: write min:0x%02x hour:0x%02x wday:0x%02x\n", 
		__func__, alarmvals[0+1], alarmvals[1+1], alarmvals[2+1]);
	}
	

	//check interrupt enable and disable
	if (rx4803->ctrlreg & (RX4803_BTC_CTRL_AIE | RX4803_BTC_CTRL_UIE) ) {
		rx4803->ctrlreg &= ~(RX4803_BTC_CTRL_AIE | RX4803_BTC_CTRL_UIE);
		err = rx4803_write_reg(spi, RX4803_BTC_CTRL, rx4803->ctrlreg);
		if (err)
			return err;		
	}
			
	//write the new minute and hour values
	alarmvals[0] = RX4803_BTC_ALARM_MIN & 0x7f;
	err = spi_write(spi, alarmvals, 4);	
	if (err)
	{
		dev_err(&spi->dev, "Unable to write register #%d\n", alarmvals[0]);
		return err;
	}

	//clear Alarm Flag
	ctrl[1] &= ~RX4803_BTC_FLAG_AF;
	
	err = rx4803_write_reg(spi, RX4803_BTC_FLAG, ctrl[1]);
	if (err)
		return err;

	//re-enable interrupt if required
	if (t->enabled) {
		//set update interrupt enable
		if ( rx4803->rtc->uie_rtctimer.enabled )
			rx4803->ctrlreg |= RX4803_BTC_CTRL_UIE;	
		//set alarm interrupt enable
		if ( rx4803->rtc->aie_timer.enabled )
			rx4803->ctrlreg |= RX4803_BTC_CTRL_AIE | RX4803_BTC_CTRL_UIE;
		
		err = rx4803_write_reg(spi, RX4803_BTC_CTRL, rx4803->ctrlreg);
		if (err)
			return err;
	}

	return 0;
}

//----------------------------------------------------------------------
// rx4803_alarm_irq_enable()
// sets enables Alarm IRQ
//
//----------------------------------------------------------------------
static int rx4803_alarm_irq_enable(struct device *dev, unsigned int enabled)
{
	struct rx4803_data *rx4803 = dev_get_drvdata(dev);
	struct spi_device *spi = rx4803->spi;
	u8 flagreg;
	u8 ctrl;
	int err;
	
	//get the current ctrl settings
	ctrl = rx4803->ctrlreg;

	dev_dbg(dev, "%s: %s => 0x%02x\n", __func__, "ctrl", ctrl);
	
	if (enabled)
	{
		// set update interrupt enable
		if ( rx4803->rtc->uie_rtctimer.enabled )
			ctrl |= RX4803_BTC_CTRL_UIE; 
		// set alarm interrupt enable		
		if ( rx4803->rtc->aie_timer.enabled )
			ctrl |= (RX4803_BTC_CTRL_AIE | RX4803_BTC_CTRL_UIE); 		
	}
	else
	{
		//clear update interrupt enable
		if ( ! rx4803->rtc->uie_rtctimer.enabled )
			ctrl &= ~RX4803_BTC_CTRL_UIE; 
		//clear alarm interrupt enable
		if ( ! rx4803->rtc->aie_timer.enabled )
		{
			if ( rx4803->rtc->uie_rtctimer.enabled )
				ctrl &= ~RX4803_BTC_CTRL_AIE;  
			else
				ctrl &= ~(RX4803_BTC_CTRL_AIE | RX4803_BTC_CTRL_UIE); 
		}		
	}
	
	dev_dbg(dev, "%s: %s => 0x%02x\n", __func__, "ctrl", ctrl);

	//clear alarm flag
	err = rx4803_read_reg(spi, RX4803_BTC_FLAG, &flagreg);
	if (err <0)
		return err;
	flagreg &= ~RX4803_BTC_FLAG_AF;
	err = rx4803_write_reg(spi, RX4803_BTC_FLAG, flagreg); 
	if (err)
		return err;
	
	//update the Control register if the setting changed
	if (ctrl != rx4803->ctrlreg) {
		rx4803->ctrlreg = ctrl;
		err = rx4803_write_reg(spi, RX4803_BTC_CTRL, rx4803->ctrlreg); 
		if (err)
			return err;
	}
	
	err = rx4803_read_reg(spi, RX4803_BTC_CTRL, &ctrl); 
	if (err)
		return err;
	dev_dbg(dev, "%s: REG[0x%02x] => 0x%02x\n", __func__, RX4803_BTC_CTRL, ctrl);
	
	return 0;
}

//---------------------------------------------------------------------------
// rx4803_ioctl()
//
//---------------------------------------------------------------------------
static int rx4803_ioctl(struct device *dev, unsigned int cmd, unsigned long arg)
{
	struct rx4803_data *rx4803 = dev_get_drvdata(dev);
	struct spi_device *spi = rx4803->spi;
	//struct mutex *lock = &rx4803->rtc->ops_lock;
	int ret = 0;
	int tmp;
	void __user *argp = (void __user *)arg;
	reg_data reg;
	
	dev_dbg(dev, "%s: cmd=%x\n", __func__, cmd);

	switch (cmd) {
		case SE_RTC_REG_READ:
			if (copy_from_user(&reg, argp, sizeof(reg)))
				return -EFAULT;
			if ( reg.number < RX4803_BTC_SEC || reg.number > RX4803_EX2_EVENT_CTRL )
				return -EFAULT;
			//mutex_lock(lock);
			ret = rx4803_read_reg(spi, reg.number, &reg.value);
			//mutex_unlock(lock);
			if (! ret )				
				return copy_to_user(argp, &reg, sizeof(reg)) ? -EFAULT : 0;
			break;
			
		case SE_RTC_REG_WRITE:
			if (copy_from_user(&reg, argp, sizeof(reg)))
				return -EFAULT;
			if ( reg.number < RX4803_BTC_SEC || reg.number > RX4803_EX2_EVENT_CTRL )
				return -EFAULT;				
			//mutex_lock(lock);
			ret = rx4803_write_reg(spi, reg.number, reg.value);
			//mutex_unlock(lock);
			break;
	
		case RTC_VL_READ:
			//mutex_lock(lock);
			ret = rx4803_read_reg(spi, RX4803_BTC_FLAG, &reg.value);
			//mutex_unlock(lock);
			if (! ret)	
			{
				tmp = !!(reg.value & (RX4803_BTC_FLAG_VLF | RX4803_BTC_FLAG_VDET));
				return copy_to_user(argp, &tmp, sizeof(tmp)) ? -EFAULT : 0;
			}			
			break;			
				
		case RTC_VL_CLR:
			//mutex_lock(lock);
			ret = rx4803_read_reg(spi, RX4803_BTC_FLAG, &reg.value);
			if (! ret)
			{
				reg.value &= ~(RX4803_BTC_FLAG_VLF | RX4803_BTC_FLAG_VDET);
				ret = rx4803_write_reg(spi, RX4803_BTC_FLAG, reg.value);
			}
			//mutex_unlock(lock);
			break;
			
		default:
			return -ENOIOCTLCMD;			
	}
	
	return ret;
}

static struct rtc_class_ops rx4803_rtc_ops = {
	.read_time = rx4803_get_time,
	.set_time = rx4803_set_time,
	.read_alarm = rx4803_get_alarm,
	.set_alarm = rx4803_set_alarm,
	.alarm_irq_enable = rx4803_alarm_irq_enable,
	.ioctl = rx4803_ioctl,
};

//----------------------------------------------------------------------
// rx4803_probe()
// probe routine for the rx4803 driver
//
//----------------------------------------------------------------------
static int rx4803_probe(struct spi_device *spi)
{
	struct rx4803_data *rx4803;
	int err, need_reset = 0;
	
	dev_dbg(&spi->dev, "IRQ %d supplied\n", spi->irq);

	if ((spi->bits_per_word && spi->bits_per_word != 8)
			|| (spi->max_speed_hz > 2000000)
			|| (spi->mode != (SPI_CS_HIGH | SPI_CPOL | SPI_CPHA))){
				
		dev_dbg(&spi->dev, "SPI settings: bits_per_word: %d, max_speed_hz: %d, mode: %xh \n",
			spi->bits_per_word, spi->max_speed_hz, spi->mode);
			
		spi->bits_per_word = 8;
		spi->max_speed_hz = 2000000;
		spi->mode = SPI_CS_HIGH | SPI_CPOL | SPI_CPHA;
		
		spi_setup(spi);	
		
		dev_dbg(&spi->dev, "SPI settings changed: bits_per_word: %d, max_speed_hz: %d, mode: %xh \n",
			spi->bits_per_word, spi->max_speed_hz, spi->mode);
		//return -EINVAL;	
	}
	
	dev_dbg(&spi->dev, "IRQ %d supplied by dts\n", spi->irq);
	
			    	
	rx4803 = devm_kzalloc(&spi->dev, sizeof(struct rx4803_data), GFP_KERNEL);
	if (!rx4803) {
		dev_err(&spi->dev, "failed to alloc memory\n");
		err = -ENOMEM;
		goto errout;
	}

	rx4803->spi = spi;
	spi_set_drvdata(spi, rx4803);

	err = rx4803_init(spi, &need_reset);
	if (err)
		goto errout;

	if (need_reset) {
		struct rtc_time tm;
		rtc_time_to_tm(0, &tm);		// set to 1970/1/1
		rx4803_set_time(&spi->dev, &tm);
		dev_warn(&spi->dev, " - time set to 1970/1/1\n");
	}

	rx4803->rtc = rtc_device_register(rx4803_driver.driver.name, &spi->dev, 
										&rx4803_rtc_ops, THIS_MODULE);
	
	
	if (IS_ERR(rx4803->rtc)) {
		err = PTR_ERR(rx4803->rtc);
		dev_err(&spi->dev, "unable to register the class device\n");
		goto errout;
	}

	if (spi->irq > 0) {		
		dev_info(&spi->dev, "IRQ %d supplied\n", spi->irq);
		INIT_WORK(&rx4803->work, rx4803_work);
		err = devm_request_threaded_irq(&spi->dev,
										spi->irq, 
										NULL, 
										rx4803_irq, 
										IRQF_TRIGGER_LOW | IRQF_ONESHOT,
										"rx4803", 
										spi);

		if (err) {
			dev_err(&spi->dev, "unable to request IRQ\n");
			goto errout_reg;
		}
	}

	rx4803->rtc->irq_freq = 1;
	rx4803->rtc->max_user_freq = 1;

	return 0;

errout_reg:
	rtc_device_unregister(rx4803->rtc);

errout:
	dev_err(&spi->dev, "probing for rx4803 failed\n");
	return err;
}

//----------------------------------------------------------------------
// rx4803_remove()
// remove routine for the rx4803 driver
//      
//----------------------------------------------------------------------
static int rx4803_remove(struct spi_device *spi)
{
	struct rx4803_data *rx4803 = spi_get_drvdata(spi);
	struct mutex *lock = &rx4803->rtc->ops_lock;

	if (spi->irq > 0) {
		mutex_lock(lock);
		rx4803->exiting = 1;
		mutex_unlock(lock);

		cancel_work_sync(&rx4803->work);
	}

	rtc_device_unregister(rx4803->rtc);

	return 0;
}

static struct spi_driver rx4803_driver = {
	.driver = {
		.name = "rtc-rx4803",
		.owner = THIS_MODULE,
	},
	.probe		= rx4803_probe,
	.remove		= rx4803_remove,
	.id_table	= rx4803_id,
};

module_spi_driver(rx4803_driver);

MODULE_AUTHOR("Val Krutov <val.krutov@erd.epson.com>");
MODULE_DESCRIPTION("RX-4803 SA/LC RTC driver");
MODULE_LICENSE("GPL");
